// vvv BEGIN RTS BOILERPLATE vvv

#include <stdio.h>   // NULL, fputs, fprintf
#include <unistd.h>  // read, write
#include <stdlib.h>  // realloc, calloc
#include <stdarg.h>  // va_arg etc.
#include <string.h>  // memcpy, memmove, strlen
#include <assert.h>  // assert


// -- RTS SETTINGS --

#ifndef TRACE
#  define TRACE 0               // trace level (0=none, 1=trace calls, 2=full)
#endif
#ifndef GCTHRESHOLD
#  define GCTHRESHOLD 262144    // gc() every GCTHRESHOLD allocs
#endif


// -- DATA TYPES --

#define SZ(N) ((N)*sizeof(val_t *))

struct val;
typedef void sc_t(struct val **, struct val **);

typedef struct val {
  union { sc_t *sc; struct val *ref; };
  int nargs;  // number (still) needed
  int napp;   // number so far applied
  struct val **args;

  struct val *wnext;        // work lists / root list
  struct val *next;         // global object list
  unsigned char redir:1;    // find sc and initial args via ref
  unsigned char active:1;   // used by gc
} val_t;


// -- MEMORY MANAGEMENT --

static val_t *gol;          // global object list
static val_t *root;         // root object(s)
size_t allocd;

val_t *new(sc_t *sc, size_t nargs) {
  val_t *r = calloc(1, sizeof(val_t));
  assert(r != NULL);
  allocd++;
  r->sc = sc;
  r->nargs = nargs;
  r->next = gol; gol = r;
  return r;
}

void delete(val_t *v) {
  free(v->args);
  free(v);
}

void resize(val_t *e, size_t napp) {
  e->args = realloc(e->args, SZ(napp));
  assert(e->args != NULL || napp == 0);
}

val_t *copy(val_t *v) {
  while(v->redir) v = v->ref;
  val_t *r = new(v->sc, v->nargs);
  resize(r, v->napp);
  memcpy(r->args, v->args, SZ(v->napp));
  r->napp = v->napp;
  return r;
}

// helper: follow redirects until target object found
// may return NULL during eval
val_t *follow(val_t *v)
{
    while(v && v->redir) v = v->ref;
    return v;
}

// gc helper: resolve redirects, mark object active, add to work list
void mark(val_t **pwork, val_t **pv)
{
    val_t *v = follow(*pv);

    if(v) {
        *pv = v;            // update reference
        if(!v->active) {
            v->active = 1;
            v->wnext = *pwork; *pwork = v;
        }
    }
}

size_t gc(void) {
  val_t *work, *p, **q; size_t i;
  size_t cr=0, cg=0, cs=0, ca=0;            // counters

  if(TRACE>1) fputs("GC:", stderr);
  // mark root set active
  for(p=root; p; p=p->wnext, cr++)
    p->active = 1;
  if(TRACE>1) fprintf(stderr, " (%zu roots)", cr);
  for(work=root; work; ) {
    for(p=work, work=NULL; p; p=p->wnext) {
      if(p->redir) mark(&work, &p->ref);
      if(p->args) for(i=0; i<p->napp; i++) mark(&work, p->args+i);
    }
  }
  for(q=&gol; p=*q; cg++) {
    if(!p->active) { *q = p->next; delete(p); cs++; }
    else { q = &p->next; p->active = 0; ca++; }
  }
  assert(cg == ca + cs);
  if(TRACE>1) fprintf(stderr, " sweep %zu/%zu\n", cs, cg);
  return ca;
}


// -- EVAL AND APPLY --

val_t *eval(val_t *e) {
  if(e->redir)
    return e->ref = eval(e->ref); // follow and update
  val_t hold = { .ref = NULL, .redir = 1 };
  hold.wnext = root; root = &hold;
  while(e->napp >= e->nargs) {
    // evaluate and let 'e' redirect to the result
    size_t keep = e->napp - e->nargs;
    e->sc(&hold.ref, e->args);
    hold.ref = follow(hold.ref);
    if(keep > 0) {
      val_t *r = hold.ref = copy(hold.ref);
      resize(r, r->napp + keep);
      memcpy(r->args + r->napp, e->args + e->nargs, SZ(keep));
      r->napp += keep;
    }
    e->redir = 1;
    e->ref = hold.ref;
    free(e->args); e->args = NULL;
    e = e->ref;
  };
  assert(root == &hold); root = root->wnext;
  return e;
}


// -- REQUIRED STANDARD COMBINATORS --

static val_t sc_zero, sc_succ;
static val_t sc_nil, sc_cons;
static val_t sc_none, sc_just;
static val_t sc_unit;
static val_t sc_eputs, sc_main;


// -- DATA HELPERS --

val_t *app(val_t *fun, size_t n, ...) {
  va_list args;
  val_t *r = copy(fun);
  resize(r, r->napp + n);
  va_start(args, n);
  for(; n>0; n--)
    r->args[r->napp++] = va_arg(args, val_t *);
  va_end(args);
  return r;
}

val_t *num(int i) {
  val_t *r = &sc_zero;
  for(; i>0; i--)
    r = app(&sc_succ, 1, r);
  return r;
}

val_t *str(const char *s) {
  char c;
  const char *t;
  val_t *r = &sc_nil;
  for(t=s+strlen(s)-1; t>=s; t--)
    r = app(&sc_cons, 2, num(*t), r);
  return r;
}

unsigned int count(val_t *n) {
  unsigned int c = 0;
  for(n = eval(n); n->napp > 0; n = eval(n->args[0])) c++;
  if(TRACE>1) fprintf(stderr, "COUNT: %u (%p)\n",c,n);
  return c;
}


// -- IO PRIMITIVES --

void primfun_putbyte(val_t **r, val_t **args) {
  char c = count(args[1]);
  write(count(args[0]), &c, 1);
  *r = &sc_unit;
}
static val_t prim_putbyte = {primfun_putbyte, 3, 0, NULL};

void primfun_getbyte(val_t **r, val_t **args) {
  char c;
  int n, fd=count(args[0]);
  if(read(fd, &c, 1) > 0)
    *r = app(&sc_just, 1, num(c));
  else
    *r = &sc_none;
}
static val_t prim_getbyte = {primfun_getbyte, 2, 0, NULL};


// -- EVALUATOR ENTRY: RUNIO, SEQ, TRACE --

void runio(val_t *e) {
  e = app(e, 3, &prim_getbyte, &prim_putbyte, &sc_unit);
  e->wnext = root; root = e;
  eval(e);
  assert(root == e); root = root->wnext;
}

void primfun_seq(val_t **r, val_t **args) {
  eval(args[0]);
  *r = eval(args[1]);
}
static val_t sc_seq = {primfun_seq, 2, 0, NULL};

void primfun_trace(val_t **r, val_t **args) {
  if(TRACE) runio(app(&sc_eputs, 1, args[0]));
  *r = eval(args[1]);
}
static val_t sc_trace = {primfun_trace, 2, 0, NULL};


// -- MAIN --

int main(int argc, char **argv) {
  runio(&sc_main);
  if(TRACE>1) fprintf(stderr, "LEFT: %zu/%zu\n", gc(), allocd);
  return 0;
}

// ^^^ END RTS BOILERPLATE ^^^
